@@ -1,22 +1,21 @@
-/* $Id: decode.c,v 1.31 2006/12/12 15:33:45 jonz Exp $ */
+/* $Id: decode.c,v 1.395 2011/09/03 13:25:39 sbajic Exp $ */
 
 /*
  DSPAM
- COPYRIGHT (C) 2002-2006 JONATHAN A. ZDZIARSKI
+ COPYRIGHT (C) 2002-2012 DSPAM PROJECT
 
- This program is free software; you can redistribute it and/or
- modify it under the terms of the GNU General Public License
- as published by the Free Software Foundation; version 2
- of the License.
+ This program is free software: you can redistribute it and/or modify
+ it under the terms of the GNU Affero General Public License as
+ published by the Free Software Foundation, either version 3 of the
+ License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- GNU General Public License for more details.
+ GNU Affero General Public License for more details.
 
- You should have received a copy of the GNU General Public License
- along with this program; if not, write to the Free Software
- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
+ You should have received a copy of the GNU Affero General Public License
+ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 */
 
@@ -45,9 +44,6 @@
 #include "buffer.h"
 #include "base64.h"
 #include "libdspam.h"
-#ifdef NCORE
-#include "ncore_adp.h"
-#endif
 
 /*
  * _ds_actualize_message (const char *message)
@@ -68,17 +64,30 @@
 ds_message_t
 _ds_actualize_message (const char *message)
 {
-  char *line, *in = strdup (message), *m_in;
+  char *line = NULL;
+  char *in = NULL;
+  char *m_in = NULL;
   ds_message_part_t current_block;
   ds_header_t current_heading = NULL;
-  struct nt *boundaries = nt_create (NT_CHAR);
-  ds_message_t out = (ds_message_t) calloc (1, sizeof (struct _ds_message));
+  struct nt *boundaries = NULL;
+  ds_message_t out = NULL;
   int block_position = BP_HEADER;
   int in_content = 0;
 
+  if (!message || !(*message))
+    goto MEMFAIL;
+
+  if (!(in = strdup(message)))
+    goto MEMFAIL;
+
   m_in = in;
 
-  if (!in || !boundaries || !out)
+  boundaries = nt_create (NT_CHAR);
+  if (!boundaries)
+    goto MEMFAIL;
+
+  out = (ds_message_t) calloc (1, sizeof (struct _ds_message));
+  if (!out)
     goto MEMFAIL;
 
   out->components = nt_create (NT_PTR);
@@ -86,16 +95,16 @@ _ds_actualize_message (const char *message)
     goto MEMFAIL;
 
   current_block = _ds_create_message_part ();
-  if (!current_block) 
+  if (!current_block)
     goto MEMFAIL;
 
   if (nt_add (out->components, (void *) current_block) == NULL)
-    LOG (LOG_CRIT, ERR_MEM_ALLOC);
+    goto MEMFAIL;
 
   /* Read the message from memory */
 
   line = strsep (&in, "\n");
-  while (line) 
+  while (line)
   {
 
     /* Header processing */
@@ -116,11 +125,8 @@ _ds_actualize_message (const char *message)
         _ds_decode_headers(current_block);
         current_block = _ds_create_message_part ();
 
-        if (!current_block) 
-        {
-          LOG (LOG_CRIT, ERR_MEM_ALLOC);
+        if (!current_block)
           goto MEMFAIL;
-        }
 
         if (nt_add (out->components, (void *) current_block) == NULL)
           goto MEMFAIL;
@@ -138,7 +144,7 @@ _ds_actualize_message (const char *message)
 
           ptr = realloc (current_heading->data,
                          strlen (current_heading->data) + strlen (line) + 2);
-          if (ptr) 
+          if (ptr)
           {
             current_heading->data = ptr;
             strcat (current_heading->data, "\n");
@@ -153,7 +159,7 @@ _ds_actualize_message (const char *message)
           ptr =
             realloc (current_heading->concatenated_data,
               strlen (current_heading->concatenated_data) + strlen (eow) + 1);
-          if (ptr) 
+          if (ptr)
           {
             current_heading->concatenated_data = ptr;
             strcat (current_heading->concatenated_data, eow);
@@ -179,9 +185,9 @@ _ds_actualize_message (const char *message)
         }
       }
 
-      /* New header field */
+      /* New header field when LF or CRLF is not found */
 
-      else if (line[0] != 0)
+      else if (line[0] != 0  && line[0] != 13)
       {
         ds_header_t header = _ds_create_header_field (line);
 
@@ -193,7 +199,7 @@ _ds_actualize_message (const char *message)
         }
 
 
-      /* line[0] == 0; switch to body */
+      /* line[0] == 0 or line[0] == 13; LF or CRLF, switch to body */
 
       } else {
         block_position = BP_BODY;
@@ -268,9 +274,9 @@ _ds_actualize_message (const char *message)
   return out;
 
 MEMFAIL:
-  free(m_in);
-  nt_destroy (boundaries);
-  _ds_destroy_message(out);
+  if (m_in) free(m_in);
+  if (boundaries) nt_destroy (boundaries);
+  if (out) _ds_destroy_message(out);
   LOG (LOG_CRIT, ERR_MEM_ALLOC);
   return NULL;
 }
@@ -289,14 +295,14 @@ MEMFAIL:
 ds_message_part_t
 _ds_create_message_part (void)
 {
-  ds_message_part_t block = 
+  ds_message_part_t block =
     (ds_message_part_t) calloc (1, sizeof (struct _ds_message_part));
 
-  if (!block) 
+  if (!block)
     goto MEMFAIL;
 
   block->headers = nt_create (NT_PTR);
-  if (!block->headers) 
+  if (!block->headers)
     goto MEMFAIL;
 
   block->body = buffer_create (NULL);
@@ -349,7 +355,7 @@ _ds_create_header_field (const char *heading)
   ds_header_t header =
     (ds_header_t) calloc (1, sizeof (struct _ds_header_field));
 
-  if (!header || !in) 
+  if (!header || !in)
     goto MEMFAIL;
 
   ptr = strsep (&in, ":");
@@ -371,8 +377,8 @@ _ds_create_header_field (const char *heading)
       else
       {
         /* Skip white space */
-        while (*in == 32 || *in == 9) 
-          ++in; 
+        while (*in == 32 || *in == 9)
+          ++in;
       }
 
       data = strdup (in);
@@ -406,6 +412,9 @@ MEMFAIL:
 
 int
 _ds_decode_headers (ds_message_part_t block) {
+#ifdef VERBOSE
+  LOGDEBUG("decoding headers in message block");
+#endif
   char *ptr, *dptr, *rest, *enc;
   ds_header_t header;
   struct nt_node *node_nt;
@@ -431,23 +440,27 @@ _ds_decode_headers (ds_message_part_t block) {
           was_null = 1;
         }
 
-        ptr = strtok_r (enc, "?", &ptrptr);
-        ptr = strtok_r (NULL, "?", &ptrptr);
+        strtok_r (enc, "?", &ptrptr);
+        strtok_r (NULL, "?", &ptrptr);
         ptr = strtok_r (NULL, "?", &ptrptr);
         dptr = strtok_r (NULL, "?", &ptrptr);
         if (!dptr) {
-          if (was_null) 
+          if (was_null && header->original_data != NULL)
+            free(header->original_data);
+          if (was_null)
             header->original_data = NULL;
           continue;
         }
 
-        rest = dptr + strlen (dptr) + 1;
-        if (rest[0]!=0)
+        rest = dptr + strlen (dptr);
+        if (rest[0]!=0) {
           rest++;
+          if (rest[0]!=0) rest++;
+        }
 
         if (ptr != NULL && (ptr[0] == 'b' || ptr[0] == 'B'))
           decoded = _ds_decode_base64 (dptr);
-        else if (ptr != NULL && (ptr[0] == 'q' || ptr[0] == 'Q')) 
+        else if (ptr != NULL && (ptr[0] == 'q' || ptr[0] == 'Q'))
           decoded = _ds_decode_quoted (dptr);
 
         decoded_len = 0;
@@ -470,16 +483,20 @@ _ds_decode_headers (ds_message_part_t block) {
 
             strcat(new_alloc, decoded);
             strcat(new_alloc, rest);
-            free(decoded); 
+            free(decoded);
             decoded = new_alloc;
           }
         }
 
-        if (decoded) { 
+        if (decoded) {
           enc_offset += (decoded_len-1);
           free(header->concatenated_data);
           header->concatenated_data = decoded;
         }
+        else if (was_null && header->original_data) {
+          free(header->original_data);
+          header->original_data = NULL;
+        }
         else if (was_null) {
           header->original_data = NULL;
         }
@@ -513,12 +530,12 @@ _ds_decode_headers (ds_message_part_t block) {
  * INPUT ARGUMENTS
  *      block		the message block to which the header belongs
  *      header		the header to analyze
- *      boundaries	a list of known boundaries found within the block 
+ *      boundaries	a list of known boundaries found within the block
  */
 
 void
 _ds_analyze_header (
-  ds_message_part_t block, 
+  ds_message_part_t block,
   ds_header_t header,
   struct nt *boundaries)
 {
@@ -613,7 +630,7 @@ _ds_analyze_header (
 
   if (!strcasecmp (header->heading, "Content-Disposition"))
   {
-    if (!strncasecmp (header->data, "inline", 6)) 
+    if (!strncasecmp (header->data, "inline", 6))
       block->content_disposition = PCD_INLINE;
     else if (!strncasecmp (header->data, "attachment", 10))
       block->content_disposition = PCD_ATTACHMENT;
@@ -639,7 +656,6 @@ _ds_destroy_message (ds_message_t message)
 {
   struct nt_node *node_nt;
   struct nt_c c;
-  int i = 0;
 
   if (message == NULL)
     return;
@@ -651,7 +667,6 @@ _ds_destroy_message (ds_message_t message)
       ds_message_part_t block = (ds_message_part_t) node_nt->ptr;
       _ds_destroy_block(block);
       node_nt = c_nt_next (message->components, &c);
-      i++;
     }
     nt_destroy (message->components);
   }
@@ -713,7 +728,7 @@ _ds_destroy_block (ds_message_part_t block)
   if (!block)
     return;
 
-  if (block->headers) 
+  if (block->headers)
   {
     _ds_destroy_headers (block);
     nt_destroy (block->headers);
@@ -728,15 +743,15 @@ _ds_destroy_block (ds_message_part_t block)
 
 /*
  * _ds_decode_block (ds_message_part_t block)
- *   
+ *
  * DESCRIPTION
  *   decodes a message block
- * 
+ *
  * INPUT ARGUMENTS
  *   block   the message block to decode
  *
  * RETURN VALUES
- *   a pointer to the allocated character array containing the decoded message 
+ *   a pointer to the allocated character array containing the decoded message
  *   NULL on failure
  */
 
@@ -754,12 +769,12 @@ _ds_decode_block (ds_message_part_t block)
 }
 
 /*
- * _ds_decode_{base64,quoted}
+ * _ds_decode_{base64,quoted,hex8bit}
  *
  * DESCRIPTION
  *   supporting block decoder functions
  *   these function call (or perform) specific decoding functions
- * 
+ *
  * INPUT ARGUMENTS
  *   body	encoded message body
  *
@@ -767,7 +782,6 @@ _ds_decode_block (ds_message_part_t block)
  *   a pointer to the allocated character array containing the decoded body
  */
 
-#ifndef NCORE
 char *
 _ds_decode_base64 (const char *body)
 {
@@ -780,65 +794,86 @@ _ds_decode_base64 (const char *body)
 char *
 _ds_decode_quoted (const char *body)
 {
-  char *out, *x;
-  char hex[3];
-  int val;
-  size_t len;
-
+#ifdef VERBOSE
+  LOGDEBUG("decoding Quoted Printable encoded buffer");
+#endif
   if (!body)
     return NULL;
 
-  out = strdup (body);
-  if (!out)
-  {
+  char *n, *out;
+  const char *end, *p;
+
+  n = out = malloc(strlen(body)+1);
+  end = body + strlen(body);
+
+  if (out == NULL) {
     LOG (LOG_CRIT, ERR_MEM_ALLOC);
     return NULL;
   }
-  len = strlen(out) + 1;
 
-  hex[2] = 0;
-  x = strchr (out, '=');
-  while (x != NULL)
-  {
-    hex[0] = x[1];
-    hex[1] = x[2];
-    if (x[1] == '\n')
-    {
-      memmove(x, x+2, len-((x+2)-out)); //strlen(x+2)+1);
-      len -= 2;
-      x = strchr (x, '=');
-    }
-    else
-    {
-      if (((hex[0] >= 'A' && hex[0] <= 'F')
-           || (hex[0] >= 'a' && hex[0] <= 'f')
-           || (hex[0] >= '0' && hex[0] <= '9'))
-          && ((hex[1] >= 'A' && hex[1] <= 'F')
-              || (hex[1] >= 'a' && hex[1] <= 'f')
-              || (hex[1] >= '0' && hex[1] <= '9')))
-      {
-        val = (int) strtol (hex, NULL, 16);
-        if (val) {
-          x[0] = val;
-          memmove(x+1, x+3, len-((x+3)-out)); //strlen(x+3)+1);
-          len -= 2;
-        }
-      }
-      x = strchr (x + 1, '=');
-    }
+  for (p = body; p < end; p++, n++) {
+    if (*p == '=') {
+      if (p[1] == '\r' && p[2] == '\n') {
+        n -= 1;
+        p += 2;
+      } else if (p[1] == '\n') {
+        n -= 1;
+        p += 1;
+      } else if (p[1] && p[2] && isxdigit((unsigned char) p[1]) && isxdigit((unsigned char) p[2])) {
+        *n = ((_ds_hex2dec((unsigned char) p[1])) << 4) | (_ds_hex2dec((unsigned char) p[2]));
+        p += 2;
+      } else
+        *n = *p;
+    } else
+      *n = *p;
   }
 
-  return out;
+  *n = '\0';
+  return (char *)out;
+}
+
+char *
+_ds_decode_hex8bit (const char *body)
+{
+#ifdef VERBOSE
+  LOGDEBUG("decoding hexadecimal 8-bit encodings in message block");
+#endif
+  if (!body)
+    return NULL;
+
+  char *n, *out;
+  const char *end, *p;
+
+  n = out = malloc(strlen(body)+1);
+  end = body + strlen(body);
+
+  if (out == NULL) {
+    LOG (LOG_CRIT, ERR_MEM_ALLOC);
+    return NULL;
+  }
+
+  for (p = body; p < end; p++, n++) {
+    if (*p == '%')
+      if (p[1] && p[2] && isxdigit((unsigned char) p[1]) && isxdigit((unsigned char) p[2])) {
+        *n = ((_ds_hex2dec((unsigned char) p[1])) << 4) | (_ds_hex2dec((unsigned char) p[2]));
+        p += 2;
+      } else
+        *n = *p;
+    else
+      *n = *p;
+  }
+
+  *n = '\0';
+  return (char *)out;
 }
-#endif /* NCORE */
 
 /*
  * _ds_encode_block (ds_message_part_t block, int encoding)
- *   
+ *
  * DESCRIPTION
  *   encodes a message block using the encoding specified and replaces the
  *   block's message body with the encoded data
- * 
+ *
  * INPUT ARGUMENTS
  *      block       the message block to encode
  *      encoding    encoding to use (EN_)
@@ -919,7 +954,9 @@ _ds_assemble_message (ds_message_t message, const char *newline)
   struct nt_c c_nt, c_nt2;
   char *heading;
   char *copyback;
+#ifdef VERBOSE
   int i = 0;
+#endif
 
   if (!out) {
     LOG (LOG_CRIT, ERR_MEM_ALLOC);
@@ -950,14 +987,14 @@ _ds_assemble_message (ds_message_t message, const char *newline)
                current_header->original_data;
 
         heading = malloc(
-            ((current_header->heading) ? strlen(current_header->heading) : 0) 
+            ((current_header->heading) ? strlen(current_header->heading) : 0)
           + ((data) ? strlen(data) : 0)
           + 3 + strlen(newline));
 
         if (current_header->heading != NULL &&
-            (!strncmp (current_header->heading, "From ", 5) || 
+            (!strncmp (current_header->heading, "From ", 5) ||
              !strncmp (current_header->heading, "--", 2)))
-          sprintf (heading, "%s:%s%s", 
+          sprintf (heading, "%s:%s%s",
             (current_header->heading) ? current_header->heading : "",
             (data) ? data : "", newline);
         else
@@ -985,9 +1022,11 @@ _ds_assemble_message (ds_message_t message, const char *newline)
       buffer_cat (out, "--");
       buffer_cat (out, block->terminating_boundary);
     }
-   
+
     node_nt = c_nt_next (message->components, &c_nt);
+#ifdef VERBOSE
     i++;
+#endif
 
     if (node_nt != NULL && node_nt->ptr != NULL)
       buffer_cat (out, newline);
@@ -1000,7 +1039,7 @@ _ds_assemble_message (ds_message_t message, const char *newline)
 }
 
 /*
- * _ds_{push,pop,match,extract}_boundary 
+ * _ds_{push,pop,match,extract}_boundary
  *
  * DESCRIPTION
  *   these functions maintain and service a boundary "stack" on the message
@@ -1031,7 +1070,7 @@ _ds_pop_boundary (struct nt *stack)
   struct nt_node *node, *last_node = NULL, *parent_node = NULL;
   struct nt_c c;
   char *boundary = NULL;
-                                                                                
+
   node = c_nt_first (stack, &c);
   while (node != NULL)
   {
@@ -1043,24 +1082,24 @@ _ds_pop_boundary (struct nt *stack)
     parent_node->next = NULL;
   else
     stack->first = NULL;
-                                                                                
+
   if (last_node == NULL)
     return NULL;
-                                                                                
+
   boundary = strdup (last_node->ptr);
-                                                                                
+
   free (last_node->ptr);
   free (last_node);
-                                                                                
+
   return boundary;
 }
-                                                                                
+
 int
 _ds_match_boundary (struct nt *stack, const char *buff)
 {
   struct nt_node *node;
   struct nt_c c;
-                                                                                
+
   node = c_nt_first (stack, &c);
   while (node != NULL)
   {
@@ -1111,26 +1150,23 @@ _ds_extract_boundary (char *buf, size_t size, char *mem)
 }
 
 /*
- * _ds_find_header (ds_message_t message, consr char *heading, int flags) {
+ * _ds_find_header (ds_message_t message, consr char *heading) {
  *
  * DESCRIPTION
  *   finds a header and returns its value
  *
  * INPUT ARGUMENTS
  *   message     the message structure to search
- *   heading	the heading to search for 	
+ *   heading	the heading to search for
  *   flags	optional search flags
  *
- * FLAGS
- *   DDF_ICASE	case insensitive search
- *
  * RETURN VALUES
  *   a pointer to the header structure's value
  *
  */
 
 char *
-_ds_find_header (ds_message_t message, const char *heading, int flags) {
+_ds_find_header (ds_message_t message, const char *heading) {
   ds_message_part_t block;
   ds_header_t head;
   struct nt_node *node_nt;
@@ -1147,12 +1183,8 @@ _ds_find_header (ds_message_t message, const char *heading, int flags) {
   node_nt = block->headers->first;
   while(node_nt != NULL) {
     head = (ds_header_t) node_nt->ptr;
-    if (flags & DDF_ICASE) {
-      if (head && !strcasecmp(head->heading, heading))
-        return head->data;
-    } else {
-      if (head && !strcmp(head->heading, heading))
-        return head->data;
+    if (head && !strcasecmp(head->heading, heading)) {
+      return head->data;
     }
     node_nt = node_nt->next;
   }
@@ -1160,3 +1192,361 @@ _ds_find_header (ds_message_t message, const char *heading, int flags) {
   return NULL;
 }
 
+int _ds_hex2dec(unsigned char hex) {
+  switch (hex) {
+    case '0': return 0;
+    case '1': return 1;
+    case '2': return 2;
+    case '3': return 3;
+    case '4': return 4;
+    case '5': return 5;
+    case '6': return 6;
+    case '7': return 7;
+    case '8': return 8;
+    case '9': return 9;
+    case 'a': case 'A': return 10;
+    case 'b': case 'B': return 11;
+    case 'c': case 'C': return 12;
+    case 'd': case 'D': return 13;
+    case 'e': case 'E': return 14;
+    case 'f': case 'F': return 15;
+    default: return -1;
+  }
+}
+
+/*
+ * _ds_strip_html(const char *html)
+ *
+ * DESCRIPTION
+ *    strip html tags from the supplied message
+ *
+ * INPUT ARGUMENTS
+ *     html encoded message body
+ *
+ * RETURN VALUES
+ *   a pointer to the allocated character array containing the
+ *   stripped message
+ *
+ */
+
+char *
+_ds_strip_html (const char *html)
+{
+#ifdef VERBOSE
+  LOGDEBUG("stripping HTML tags from message block");
+#endif
+  size_t j = 0, k = 0, i = 0;
+  int visible = 1;
+  int closing_td_tag = 0;
+  char *html2;
+  const char *cdata_close_tag = NULL;
+
+  if(!html)
+    return NULL;
+
+  static struct {
+    unsigned int id;
+    char *entity;
+  }
+  charset[] = {
+    {   32, "&nbsp;"    }, {  34, "&quot;"    }, {  34, "&quot;"    }, {  38, "&amp;"     },
+    {   38, "&amp;"     }, {  39, "&apos;"    }, {  60, "&lt;"      }, {  60, "&lt;"      },
+    {   62, "&gt;"      }, {  62, "&gt;"      }, { 160, "&nbsp;"    }, { 161, "&iexcl;"   },
+    {  162, "&cent;"    }, { 163, "&pound;"   }, { 164, "&curren;"  }, { 165, "&yen;"     },
+    {  166, "&brvbar;"  }, { 167, "&sect;"    }, { 168, "&uml;"     }, { 169, "&copy;"    },
+    {  170, "&ordf;"    }, { 171, "&laquo;"   }, { 172, "&not;"     }, { 173, "&shy;"     },
+    {  174, "&reg;"     }, { 175, "&macr;"    }, { 176, "&deg;"     }, { 177, "&plusmn;"  },
+    {  178, "&sup2;"    }, { 179, "&sup3;"    }, { 180, "&acute;"   }, { 181, "&micro;"   },
+    {  182, "&para;"    }, { 183, "&middot;"  }, { 184, "&cedil;"   }, { 185, "&sup1;"    },
+    {  186, "&ordm;"    }, { 187, "&raquo;"   }, { 188, "&frac14;"  }, { 189, "&frac12;"  },
+    {  190, "&frac34;"  }, { 191, "&iquest;"  }, { 192, "&Agrave;"  }, { 193, "&Aacute;"  },
+    {  194, "&Acirc;"   }, { 195, "&Atilde;"  }, { 196, "&Auml;"    }, { 197, "&Aring;"   },
+    {  198, "&AElig;"   }, { 199, "&Ccedil;"  }, { 200, "&Egrave;"  }, { 201, "&Eacute;"  },
+    {  202, "&Ecirc;"   }, { 203, "&Euml;"    }, { 204, "&Igrave;"  }, { 205, "&Iacute;"  },
+    {  206, "&Icirc;"   }, { 207, "&Iuml;"    }, { 208, "&ETH;"     }, { 209, "&Ntilde;"  },
+    {  210, "&Ograve;"  }, { 211, "&Oacute;"  }, { 212, "&Ocirc;"   }, { 213, "&Otilde;"  },
+    {  214, "&Ouml;"    }, { 215, "&times;"   }, { 216, "&Oslash;"  }, { 217, "&Ugrave;"  },
+    {  218, "&Uacute;"  }, { 219, "&Ucirc;"   }, { 220, "&Uuml;"    }, { 221, "&Yacute;"  },
+    {  222, "&THORN;"   }, { 223, "&szlig;"   }, { 224, "&agrave;"  }, { 225, "&aacute;"  },
+    {  226, "&acirc;"   }, { 227, "&atilde;"  }, { 228, "&auml;"    }, { 229, "&aring;"   },
+    {  230, "&aelig;"   }, { 231, "&ccedil;"  }, { 232, "&egrave;"  }, { 233, "&eacute;"  },
+    {  234, "&ecirc;"   }, { 235, "&euml;"    }, { 236, "&igrave;"  }, { 237, "&iacute;"  },
+    {  238, "&icirc;"   }, { 239, "&iuml;"    }, { 240, "&eth;"     }, { 241, "&ntilde;"  },
+    {  242, "&ograve;"  }, { 243, "&oacute;"  }, { 244, "&ocirc;"   }, { 245, "&otilde;"  },
+    {  246, "&ouml;"    }, { 247, "&divide;"  }, { 248, "&oslash;"  }, { 249, "&ugrave;"  },
+    {  250, "&uacute;"  }, { 251, "&ucirc;"   }, { 252, "&uuml;"    }, { 253, "&yacute;"  },
+    {  254, "&thorn;"   }, { 255, "&yuml;"    }, { 338, "&OElig;"   }, { 339, "&oelig;"   },
+    {  352, "&Scaron;"  }, { 353, "&scaron;"  }, { 376, "&Yuml;"    }, { 402, "&fnof;"    },
+    {  710, "&circ;"    }, { 732, "&tilde;"   }, { 913, "&Alpha;"   }, { 914, "&Beta;"    },
+    {  915, "&Gamma;"   }, { 916, "&Delta;"   }, { 917, "&Epsilon;" }, { 918, "&Zeta;"    },
+    {  919, "&Eta;"     }, { 920, "&Theta;"   }, { 921, "&Iota;"    }, { 922, "&Kappa;"   },
+    {  923, "&Lambda;"  }, { 924, "&Mu;"      }, { 925, "&Nu;"      }, { 926, "&Xi;"      },
+    {  927, "&Omicron;" }, { 928, "&Pi;"      }, { 929, "&Rho;"     }, { 931, "&Sigma;"   },
+    {  932, "&Tau;"     }, { 933, "&Upsilon;" }, { 934, "&Phi;"     }, { 935, "&Chi;"     },
+    {  936, "&Psi;"     }, { 937, "&Omega;"   }, { 945, "&alpha;"   }, { 946, "&beta;"    },
+    {  947, "&gamma;"   }, { 948, "&delta;"   }, { 949, "&epsilon;" }, { 950, "&zeta;"    },
+    {  951, "&eta;"     }, { 952, "&theta;"   }, { 953, "&iota;"    }, { 954, "&kappa;"   },
+    {  955, "&lambda;"  }, { 956, "&mu;"      }, { 957, "&nu;"      }, { 958, "&xi;"      },
+    {  959, "&omicron;" }, { 960, "&pi;"      }, { 961, "&rho;"     }, { 962, "&sigmaf;"  },
+    {  963, "&sigma;"   }, { 964, "&tau;"     }, { 965, "&upsilon;" }, { 966, "&phi;"     },
+    {  967, "&chi;"     }, { 968, "&psi;"     }, { 969, "&omega;"   }, { 977, "&thetasym" },
+    {  978, "&upsih;"   }, { 982, "&piv;"     }, {8194, "&ensp;"    }, {8195, "&emsp;"    },
+    { 8201, "&thinsp;"  }, {8204, "&zwnj;"    }, {8205, "&zwj;"     }, {8206, "&lrm;"     },
+    { 8207, "&rlm;"     }, {8211, "&ndash;"   }, {8212, "&mdash;"   }, {8216, "&lsquo;"   },
+    { 8217, "&rsquo;"   }, {8218, "&sbquo;"   }, {8220, "&ldquo;"   }, {8221, "&rdquo;"   },
+    { 8222, "&bdquo;"   }, {8224, "&dagger;"  }, {8225, "&Dagger;"  }, {8226, "&bull;"    },
+    { 8230, "&hellip;"  }, {8240, "&permil;"  }, {8242, "&prime;"   }, {8243, "&Prime;"   },
+    { 8249, "&lsaquo;"  }, {8250, "&rsaquo;"  }, {8254, "&oline;"   }, {8260, "&frasl;"   },
+    { 8364, "&euro;"    }, {8465, "&image;"   }, {8472, "&weierp;"  }, {8476, "&real;"    },
+    { 8482, "&trade;"   }, {8501, "&alefsym;" }, {8592, "&larr;"    }, {8593, "&uarr;"    },
+    { 8594, "&rarr;"    }, {8595, "&darr;"    }, {8596, "&harr;"    }, {8629, "&crarr;"   },
+    { 8656, "&lArr;"    }, {8657, "&uArr;"    }, {8658, "&rArr;"    }, {8659, "&dArr;"    },
+    { 8660, "&hArr;"    }, {8704, "&forall;"  }, {8706, "&part;"    }, {8707, "&exist;"   },
+    { 8709, "&empty;"   }, {8711, "&nabla;"   }, {8712, "&isin;"    }, {8713, "&notin;"   },
+    { 8715, "&ni;"      }, {8719, "&prod;"    }, {8721, "&sum;"     }, {8722, "&minus;"   },
+    { 8727, "&lowast;"  }, {8730, "&radic;"   }, {8733, "&prop;"    }, {8734, "&infin;"   },
+    { 8736, "&ang;"     }, {8743, "&and;"     }, {8744, "&or;"      }, {8745, "&cap;"     },
+    { 8746, "&cup;"     }, {8747, "&int;"     }, {8756, "&there4;"  }, {8764, "&sim;"     },
+    { 8773, "&cong;"    }, {8776, "&asymp;"   }, {8800, "&ne;"      }, {8801, "&equiv;"   },
+    { 8804, "&le;"      }, {8805, "&ge;"      }, {8834, "&sub;"     }, {8835, "&sup;"     },
+    { 8836, "&nsub;"    }, {8838, "&sube;"    }, {8839, "&supe;"    }, {8853, "&oplus;"   },
+    { 8855, "&otimes;"  }, {8869, "&perp;"    }, {8901, "&sdot;"    }, {8968, "&lceil;"   },
+    { 8969, "&rceil;"   }, {8970, "&lfloor;"  }, {8971, "&rfloor;"  }, {9001, "&lang;"    },
+    { 9002, "&rang;"    }, {9674, "&loz;"     }, {9824, "&spades;"  }, {9827, "&clubs;"   },
+    { 9829, "&hearts;"  }, {9830, "&diams;"   }
+  };
+  int num_chars = sizeof(charset) / sizeof(charset[0]);
+
+  static struct {
+    char *open_tag;
+    char *uri_tag;
+  }
+  uritag[] = {
+    {          "<a", "href"       }, {        "<img", "src"        }, {      "<input", "src"        },
+    {     "<iframe", "src"        }, {      "<frame", "src"        }, {     "<script", "src"        },
+    {       "<form", "action"     }, {      "<embed", "src"        }, {       "<area", "href"       },
+    {       "<base", "href"       }, {       "<link", "href"       }, {     "<source", "src"        },
+    {       "<body", "background" }, { "<blockquote", "cite"       }, {          "<q", "cite"       },
+    {        "<ins", "cite"       }, {        "<del", "cite"       }
+  };
+  int num_uri = sizeof(uritag) / sizeof(uritag[0]);
+
+  size_t len = strlen(html);
+  html2 = malloc(len+1);
+
+  if (html2 == NULL) {
+    LOG (LOG_CRIT, ERR_MEM_ALLOC);
+    return NULL;
+  }
+
+  for (i = 0; i < len; i++) {
+    if (html[i] == '<') {
+      if (cdata_close_tag) {
+        if (strncasecmp(html + i, cdata_close_tag, strlen(cdata_close_tag)) == 0) {
+          i += strlen(cdata_close_tag) - 1;
+          cdata_close_tag = NULL;
+        }
+        continue;
+      } else if (strncasecmp(html + i, "</td>", 5) == 0) {
+        i += 4;
+        closing_td_tag = 1;
+        continue;
+      } else if (strncasecmp(html + i, "<td", 3) == 0 && closing_td_tag) {
+        if (j > 0 && !isspace(html2[j-1])) {
+          html2[j++]=' ';
+        }
+        visible = 0;
+      } else {
+        closing_td_tag = 0;
+        visible = 1;
+      }
+      k = i + 1;
+
+      if ((k < len) && (!( (html[k] >= 65 && html[k] <= 90) ||
+                           (html[k] >= 97 && html[k] <= 122) ||
+                           (html[k] == 47) ||
+                           (html[k] == 33) ))) {
+        /* Not a HTML tag. HTML tags start with a letter, forwardslash or exclamation mark */
+        visible = 1;
+        html2[j++]=html[i];
+        i = k;
+        const char *w = &(html[k]);
+        while (j < len && (size_t)(w - html) < len && *w != '<') {
+          html2[j++]=*w;
+          w++;
+          i++;
+        }
+        continue;
+      } else if (html[k]) {
+        /* find the end of the tag */
+        while (k < len && html[k] != '<' && html[k] != '>') {k++;}
+
+        /* if we've got a tag with a uri, save the address to print later. */
+        char *url_tag = " ";
+        int tag_offset = 0, x = 0, y = 0;
+        for (y = 0; y < num_uri; y++) {
+          x = strlen(uritag[y].open_tag);
+          if (strncasecmp(html+i,uritag[y].open_tag,x)==0 && (i+x < len && isspace(html[i+x]))) {
+            url_tag = uritag[y].uri_tag;
+            tag_offset = i + x + 1;
+            break;
+          }
+        }
+        /* tag with uri found */
+        if (tag_offset > 0) {
+          size_t url_start;         /* start of url tag inclusive [ */
+          size_t url_tag_len = strlen(url_tag);
+          char delim = ' ';
+          /* find start of uri */
+          for (url_start = tag_offset; url_start <= k; url_start++) {
+            if (strncasecmp(html + url_start, url_tag, url_tag_len) == 0) {
+              url_start += url_tag_len;
+              while (html[url_start] && isspace(html[url_start])) {url_start++;}   /* remove spaces before = */
+              if (html[url_start] == '=') {
+                url_start++;
+                while (html[url_start] && isspace(html[url_start])) {url_start++;} /* remove spaces after = */
+                if (html[url_start] == '"') {
+                  delim = '"';
+                  url_start++;
+                } else if (html[url_start] == '\'') {
+                  delim = '\'';
+                  url_start++;
+                } else {
+                  delim = '>';
+                }
+                break;
+              } else {
+                /* Start of uri tag found but no '=' after the tag.
+                 * Skip the whole tag.
+                 */
+                break;
+              }
+            } else if ((url_start - tag_offset) >= 50) {
+              /* The length of the html tag is over 50 characters long without
+               * finding the start of the url/uri. Skip the whole tag.
+               */
+              break;
+            }
+          }
+          /* find end of uri */
+          if (delim != ' ') {
+            if (url_start < len &&
+                (strncasecmp(html + url_start, "http:", 5) == 0 ||
+                 strncasecmp(html + url_start, "https:", 6) == 0 ||
+                 strncasecmp(html + url_start, "ftp:", 4) == 0)) {
+              html2[j++]=' ';
+              const char *w = &(html[url_start]);
+              /* html2 is a buffer of len + 1, where the +1 is for NULL
+               * termination. This means we only want to loop to len
+               * since we will replace html2[j] right after the loop.
+               */
+              while (j < len && (size_t)(w - html) < len && *w != delim) {
+                html2[j++]=*w;
+                w++;
+              }
+              html2[j++]=' ';
+            }
+          }
+        } else if (strncasecmp(html + i, "<p>", 3) == 0
+                || strncasecmp(html + i, "<p ", 3) == 0
+                || strncasecmp(html + i, "<p\t", 3) == 0
+                || strncasecmp(html + i, "<tr", 3) == 0
+                || strncasecmp(html + i, "<option", 7) == 0
+                || strncasecmp(html + i, "<br", 3) == 0
+                || strncasecmp(html + i, "<li", 3) == 0
+                || strncasecmp(html + i, "<div", 4) == 0
+                || strncasecmp(html + i, "</select>", 9) == 0
+                || strncasecmp(html + i, "</table>", 8) == 0) {
+          if (j > 0 && html2[j-1] != '\n' && html2[j-1] != '\r') {
+            html2[j++] = '\n';
+          }
+        } else if (strncasecmp(html + i, "<applet", 7) == 0) {
+          cdata_close_tag = "</applet>";
+        } else if (strncasecmp(html + i, "<embed", 6) == 0) {
+          cdata_close_tag = "</embed>";
+        } else if (strncasecmp(html + i, "<frameset", 9) == 0) {
+          cdata_close_tag = "</frameset>";
+        } else if (strncasecmp(html + i, "<frame", 6) == 0) {
+          cdata_close_tag = "</frame>";
+        } else if (strncasecmp(html + i, "<iframe", 7) == 0) {
+          cdata_close_tag = "</iframe>";
+        } else if (strncasecmp(html + i, "<noembed", 8) == 0) {
+          cdata_close_tag = "</noembed>";
+        } else if (strncasecmp(html + i, "<noscript", 9) == 0) {
+          cdata_close_tag = "</noscript>";
+        } else if (strncasecmp(html + i, "<object", 7) == 0) {
+          cdata_close_tag = "</object>";
+        } else if (strncasecmp(html + i, "<script", 7) == 0) {
+          cdata_close_tag = "</script>";
+        } else if (strncasecmp(html + i, "<style", 6) == 0) {
+          cdata_close_tag = "</style>";
+        }
+        i = (html[k] == '<' || html[k] == '\0')? k - 1: k;
+        continue;
+      }
+    } else if (cdata_close_tag) {
+      continue;
+    } else if (!isspace(html[i])) {
+      visible = 1;
+    }
+
+    if (strncmp(html+i,"&#",2)==0) {
+      int x = 0;
+      const char *w = &(html[i+2]);
+      while (*w == '0') {i++;w++;}
+      char n[5];
+      if (html[i+4] && html[i+4] == ';'
+          && isdigit(html[i+2])
+          && isdigit(html[i+3])) {
+        n[0] = html[i+2];
+        n[1] = html[i+3];
+        n[2] = 0;
+        x = atoi(n);
+        if (x <= 255 && x >= 32)
+          html2[j++] = x;
+        i += 4;
+      } else if (html[i+6]
+                  && html[i+6] == ';'
+                  && isdigit(html[i+2])
+                  && isdigit(html[i+3])
+                  && isdigit(html[i+4])
+                  && isdigit(html[i+5])) {
+        n[0] = html[i+2];
+        n[1] = html[i+3];
+        n[2] = html[i+4];
+        n[3] = html[i+5];
+        n[4] = 0;
+        x = atoi(n);
+        if (x <= 255 && x >= 32)
+          html2[j++] = x;
+        i += 6;
+      } else {
+        const char *w = &(html[i]);
+        while (*w != ';' && *w != ' ' && *w != '\t' && *w != '\0') {i++;w++;}
+      }
+      visible = 0;
+      continue;
+    } else if (html[i] == '&') {
+      int x = 0, y = 0;
+      for (y = 0; y < num_chars; y++) {
+        x = strlen(charset[y].entity);
+        if (strncasecmp(html+i,charset[y].entity,x)==0) {
+          if (charset[y].id <= 255)
+            html2[j++] = charset[y].id;
+          i += x-1;
+          visible = 0;
+          continue;
+        }
+      }
+    }
+
+    if (j < len && visible)
+      html2[j++] = html[i];
+
+    if (j >= len)
+      i = j = len;
+  }
+
+  html2[j] = '\0';
+  return (char *)html2;
+}
